/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.engine.cache;

import com.funambol.exchange.DataAccessException;
import com.funambol.exchange.admin.ExchangeConnectorConfig;
import com.funambol.exchange.items.common.manager.ItemManager;
import com.funambol.exchange.items.common.model.Item;
import com.funambol.exchange.util.Def;
import com.funambol.framework.core.AlertCode;
import com.funambol.framework.engine.SyncItem;
import com.funambol.framework.engine.SyncItemKey;
import com.funambol.framework.engine.SyncItemState;
import com.funambol.framework.engine.source.SyncContext;
import com.funambol.framework.engine.source.SyncSource;
import com.funambol.framework.engine.source.SyncSourceException;
import com.funambol.framework.engine.source.AbstractSyncSource;
import com.funambol.framework.logging.Sync4jLogger;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * <p>This class represents a SyncSource that read and store the state of the items in a support table.</p>
 */
public abstract class ExchangeCachedSyncSource
        extends AbstractSyncSource
        implements java.io.Serializable {
    // --------------------------------------------------------------- Constants

    // ------------------------------------------------------------ Private data

    private ItemManager im = null;

    private Item[] newItems = null;
    private Item[] deleteItems = null;
    private Item[] updateItems = null;

    private ArrayList exchangeItemsKeys = null;

    private ArrayList cacheSyncItems = null;

    private ArrayList syncNewItems = null;
    private ArrayList syncUpdateItems = null;
    private ArrayList syncRemoveItems = null;

    private ArrayList syncErrorNewItems = null;
    private ArrayList syncErrorUpdateItems = null;
    private ArrayList syncErrorRemoveItems = null;

    protected Logger log = null;

    protected long principalId = 0;
    protected String username = null;
    protected String credentials = null;
    protected int syncMode = 0;

    // -------------------------------------------------------------- Properties

    /**
     * Property exchange folder
     */
    protected String exchangeFolder;

    /**
     * Getter for property exchange folder.
     *
     * @return Value of property exchange folder.
     */
    public String getExchangeFolder() {
        return exchangeFolder;
    }

    /**
     * Setter for property exchange folder.
     *
     * @param exchangeFolder New value of property exchange folder.
     */
    public void setExchangeFolder(String exchangeFolder) {
        this.exchangeFolder = exchangeFolder;
    }

    /**
     * Property SECC host
     */
    protected String host;

    /**
     * Property SECC Port
     */
    protected int port = 80;

    /**
     * Property Exchange host
     */
    protected String exchangeHost;

    // ---------------------------------------------------------- Public Methods

    /**
     * <p>Method by the server at the begining of the sync process.</p>
     *
     * @param context the context of the sync
     * @throws SyncSourceException if an error occurs
     * @see SyncSource
     */
    public void beginSync(SyncContext context)
            throws SyncSourceException {

        Item[] localItems;
        Item[] exchangeItems;

        log = Sync4jLogger.getLogger(Def.LOGGER_NAME);

        syncNewItems = new ArrayList();
        syncUpdateItems = new ArrayList();
        syncRemoveItems = new ArrayList();

        exchangeItemsKeys = new ArrayList();

        cacheSyncItems = new ArrayList();

        syncErrorNewItems = new ArrayList();
        syncErrorUpdateItems = new ArrayList();
        syncErrorRemoveItems = new ArrayList();

        try {

            this.host = ExchangeConnectorConfig.getConfigInstance().getSeccServer();
            this.port = ExchangeConnectorConfig.getConfigInstance().getSeccPort();
            this.exchangeHost = ExchangeConnectorConfig.getConfigInstance().getExchangeServer();

            initManager();

            principalId = context.getPrincipal().getId();
            username = context.getPrincipal().getUsername();
            credentials = context.getPrincipal().getEncodedCredentials();
            this.syncMode = context.getSyncMode();


            im = new ItemManager(this.host,
                    this.port,
                    this.exchangeHost);

            localItems = im.getLocalItems(this.sourceURI,
                    principalId);

            if (this.syncMode == AlertCode.REFRESH_FROM_CLIENT) {
                if (log.isLoggable(Level.FINEST)) {
                    log.log(Level.FINEST, "Using Refresh From Client Sync Method");
                }
                this.deleteAll();
                this.syncMode = AlertCode.SLOW;
                return;
            }

            //
            // in slow sync
            // in getExchangeItems() implementation
            // also load cacheSyncItems
            //
            exchangeItems = getExchangeItemsInMultipleRequests(username,
                    credentials,
                    this.exchangeFolder,
                    new String[0],
                    true);

            exchangeItemsKeys = getSyncItemKeysArrayListFromItems(exchangeItems);

            newItems = findNewItems(localItems, exchangeItems);
            updateItems = findUpdateItems(localItems, exchangeItems);
            deleteItems = findDeleteItems(localItems, exchangeItems);

            im.addItems(newItems,
                    this.sourceURI,
                    principalId);

            im.updateItems(updateItems,
                    this.sourceURI,
                    principalId);

            im.deleteItems(deleteItems,
                    this.sourceURI,
                    principalId);


            if (log.isLoggable(Level.INFO)) {
                log.info("Exchange SyncSource " + sourceURI + " - begineSync");
            }

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "beginSync", e);
            }

            throw new SyncSourceException("Error accessing items: " +
                    e.getMessage(), e);
        }

    }

    /**
     * deletes all Items from the Exchange server
     */
    public abstract void deleteAll() throws SyncSourceException;


    /**
     * @see SyncSource
     */
    public void endSync()
            throws SyncSourceException {

        String id;

        try {

            int count = 0;

            Item[] sni = (Item[]) syncNewItems.toArray(new Item[0]);

            Item[] sui = (Item[]) syncUpdateItems.toArray(new Item[0]);

            int k = sni.length + sui.length;

            //
            // upgrade only if items is added / updated in sync
            //
            if (k > 0) {

                String[] ids = new String [k];

                for (int i = 0, l = sni.length; i < l; i++) {
                    ids[count++] = sni[i].getId();
                }

                for (int i = 0, l = sui.length; i < l; i++) {
                    ids[count++] = sui[i].getId();
                }

                Item[] exchangeSyncItems = getExchangeItemsInMultipleRequests(username,
                        credentials,
                        this.exchangeFolder,
                        ids,
                        false);

                im.upgradeItems(exchangeSyncItems,
                        this.sourceURI,
                        principalId);

            }

            im.deleteItems((Item[]) syncRemoveItems.toArray(new Item[syncRemoveItems.size()]),
                    this.sourceURI,
                    principalId);

            if (log.isLoggable(Level.INFO)) {
                log.info("Exchange SyncSource " + sourceURI + " - endSync");
            }

            if (log.isLoggable(Level.FINEST)) {

                log.finest("Items added: " + syncNewItems.size());

                Iterator i = syncNewItems.iterator();

                while (i.hasNext()) {
                    id = ((Item) i.next()).getId();
                    if (log.isLoggable(Level.ALL)) {
                        log.log(Level.ALL, id);
                    }
                }

                log.finest("Items updated: " + syncUpdateItems.size());

                i = syncUpdateItems.iterator();

                while (i.hasNext()) {
                    id = ((Item) i.next()).getId();
                    if (log.isLoggable(Level.ALL)) {
                        log.log(Level.ALL, id);
                    }
                }

                log.finest("Items removed: " + syncRemoveItems.size());

                i = syncRemoveItems.iterator();

                while (i.hasNext()) {
                    id = ((Item) i.next()).getId();
                    if (log.isLoggable(Level.ALL)) {
                        log.log(Level.ALL, id);
                    }
                }

            }

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "endSync", e);
            }

            throw new SyncSourceException("Error end sync:" +
                    e.getMessage(), e);
        }

    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getNewSyncItemKeys(Timestamp since,
                                            Timestamp until)
            throws SyncSourceException {

        SyncItemKey[] syncItemKeys;
        String     [] ids;

        try {

            if (newItems != null && newItems.length > 0) {

                int l = newItems.length;

                syncItemKeys = new SyncItemKey [l];
                ids = new String      [l];

                for (int i = 0; i < l; i++) {
                    syncItemKeys[i] = new SyncItemKey(newItems[i].getId());
                    ids[i] = newItems[i].getId();
                }

                addToCacheSyncItems(getSyncItems(username,
                        credentials,
                        ids,
                        this.exchangeFolder));

            } else {
                syncItemKeys = new SyncItemKey[0];
            }

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "getNewSyncItemKeys", e);
            }

            throw new SyncSourceException(e);

        }

        return syncItemKeys;
    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getUpdatedSyncItemKeys(Timestamp since,
                                                Timestamp until)
            throws SyncSourceException {

        SyncItemKey[] syncItemKeys;
        String     [] ids;

        try {

            if (updateItems != null && updateItems.length > 0) {

                int l = updateItems.length;

                syncItemKeys = new SyncItemKey [l];
                ids = new String      [l];

                for (int i = 0; i < l; i++) {
                    syncItemKeys[i] = new SyncItemKey(updateItems[i].getId());
                    ids[i] = updateItems[i].getId();
                }

                addToCacheSyncItems(getSyncItems(username,
                        credentials,
                        ids,
                        this.exchangeFolder
                ));

            } else {
                syncItemKeys = new SyncItemKey[0];
            }

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "getUpdatedSyncItemKeys", e);
            }

            throw new SyncSourceException(e);

        }


        return syncItemKeys;
    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getDeletedSyncItemKeys(Timestamp since,
                                                Timestamp until)
            throws SyncSourceException {

        SyncItemKey[] syncItemKeys;

        try {

            if (deleteItems != null && deleteItems.length > 0) {

                int l = deleteItems.length;

                syncItemKeys = new SyncItemKey [l];

                for (int i = 0; i < l; i++) {
                    syncItemKeys[i] = new SyncItemKey(deleteItems[i].getId());
                }

                /**
                 addToCacheSyncItems(getSyncItems(username    ,
                 credentials ,
                 ids,
                 this.exchangeFolder));
                 */

            } else {
                syncItemKeys = new SyncItemKey[0];
            }

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "getDeletedSyncItemKeys", e);
            }

            throw new SyncSourceException(e);

        }

        return syncItemKeys;
    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getAllSyncItemKeys()
            throws SyncSourceException {

        return (SyncItemKey[]) exchangeItemsKeys.toArray(new SyncItemKey[exchangeItemsKeys.size()]);

    }

    /**
     * @see SyncSource
     */
    public char getSyncItemStateFromId(SyncItemKey syncItemKey)
            throws SyncSourceException {
        return getSyncItemStatus(syncItemKey.getKeyAsString());
    }

    /**
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @param ids
     * @param isBeginSync
     * @return items array
     */
    public abstract Item[] getExchangeItems(String username,
                                            String credentials,
                                            String exchangeFolder,
                                            String[] ids,
                                            boolean isBeginSync)
            throws SyncSourceException;

    /**
     * <p> Returs some properties of the Exchange items whose ids are given as
     * parameter, to be used in the cache update. The items are retrieve from Exchange
     * in more then one requests for safety reasons.</p>
     *
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @param ids
     * @param isBeginSync
     * @return items array
     * @throws SyncSourceException
     */
    private Item[] getExchangeItemsInMultipleRequests(String username,
                                                      String credentials,
                                                      String exchangeFolder,
                                                      String[] ids,
                                                      boolean isBeginSync)
            throws SyncSourceException {
        List idsList, itemsList;
        int fullRequestsCount, remeinder;
        Item[] items;

        fullRequestsCount = ids.length / Def.MAX_IDS_COUNT_IN_WEBDAV;

        idsList = new ArrayList();

        for (int i = 0; i < ids.length; i++) {
            idsList.add(ids[i]);
        }

        itemsList = new ArrayList();

        for (int i = 0; i < fullRequestsCount; i++) {

            items = getExchangeItems(username, credentials,
                    exchangeFolder,
                    (String[]) idsList.subList(i * Def.MAX_IDS_COUNT_IN_WEBDAV,
                            (i + 1) * Def.MAX_IDS_COUNT_IN_WEBDAV).toArray(new String[0]),
                    isBeginSync);
            for (int j = 0; j < items.length; j++) {
                itemsList.add(items[j]);
            }
        }

        remeinder = ids.length % Def.MAX_IDS_COUNT_IN_WEBDAV;
        items = getExchangeItems(username, credentials,
                exchangeFolder,
                (String[]) idsList.subList(ids.length - remeinder, ids.length).toArray(new String[0]),
                isBeginSync);
        for (int j = 0; j < items.length; j++) {
            itemsList.add(items[j]);
        }

        return (Item[]) itemsList.toArray(new Item[0]);
    }

    /**
     * @param ids
     * @return items array
     */
    public abstract SyncItem[] getSyncItems(String username,
                                            String credentials,
                                            String[] ids,
                                            String exchangeFolder)
            throws SyncSourceException;


    /**
     * init manager
     */
    public abstract void initManager()
            throws SyncSourceException;

    // --------------------------------------------------------- Private Methods

    /**
     * Create new items array
     *
     * @param localItems
     * @param exchangeItems
     * @return new items array
     */
    private Item[] findNewItems(Item[] localItems, Item[] exchangeItems) {

        ArrayList newRows;
        Item exchangeItem;

        int count = 0;

        newRows = new ArrayList();

        if (exchangeItems != null) {

            for (int i = 0, l = exchangeItems.length; i < l; i++) {

                exchangeItem = exchangeItems[i];

                if (!this.isItemInArray(localItems, exchangeItem)) {
                    newRows.add(exchangeItem);
                    count++;
                }

            }

        }

        return (Item[]) newRows.toArray(new Item[count]);

    }

    /**
     * Create deleted items array
     *
     * @param localItems
     * @param exchangeItems
     * @return deleted items array
     */
    private Item[] findDeleteItems(Item[] localItems, Item[] exchangeItems) {

        ArrayList deleteRows;
        Item localItem;

        int count = 0;

        deleteRows = new ArrayList();

        if (localItems != null) {

            for (int i = 0, l = localItems.length; i < l; i++) {

                localItem = localItems[i];

                if (!this.isItemInArray(exchangeItems, localItem)) {
                    deleteRows.add(localItem);
                    count++;
                }

            }

        }

        return (Item[]) deleteRows.toArray(new Item[count]);

    }

    /**
     * Create updated items array
     *
     * @param localItems
     * @param exchangeItems
     * @return updated items array
     */
    private Item[] findUpdateItems(Item[] localItems, Item[] exchangeItems) {

        ArrayList updateRows;

        Item exchangeItem;
        Item localItem;

        Date excItemLastModified;
        Date localItemLastModified;

        String excItemLastModifiedStr;
        String localItemLastModifiedStr;

        int count = 0;

        updateRows = new ArrayList();


        if (exchangeItems != null) {

            for (int i = 0, l = exchangeItems.length; i < l; i++) {

                exchangeItem = exchangeItems[i];
                localItem = getItemById(localItems,
                        exchangeItem.getId());

                if (localItem != null) {

                    excItemLastModified = exchangeItem.getLastModified();
                    localItemLastModified = localItem.getLastModified();

                    //
                    // because java.util.Date no save millisecond
                    // string rappresention of date is in HH:mm:ss
                    //
                    excItemLastModifiedStr = String.valueOf(excItemLastModified);
                    localItemLastModifiedStr = String.valueOf(localItemLastModified);

                    if (!excItemLastModifiedStr.equals(localItemLastModifiedStr)) {

                        updateRows.add(exchangeItem);
                        count++;

                    }

                }

            }

        }

        return (Item[]) updateRows.toArray(new Item[count]);

    }

    /**
     * Return SyncItem status
     *
     * @param key SyncItem key
     */
    protected char getSyncItemStatus(String key) {

        if (newItems != null) {
            for (int i = 0, l = newItems.length; i < l; i++) {
                if (newItems[i].getId().equals(key)) {
                    return SyncItemState.NEW;
                }
            }
        }

        if (updateItems != null) {
            for (int i = 0, l = updateItems.length; i < l; i++) {
                if (updateItems[i].getId().equals(key)) {
                    return SyncItemState.UPDATED;
                }
            }
        }

        if (deleteItems != null) {
            for (int i = 0, l = deleteItems.length; i < l; i++) {
                if (deleteItems[i].getId().equals(key)) {
                    return SyncItemState.DELETED;
                }
            }
        }

        return SyncItemState.SYNCHRONIZED;

    }

    // ------------------------------------------------------- Protected Methods

    /**
     * Return href (webdav resource name) of item
     *
     * @param id        id of item to find
     * @param principal
     * @return href (webdav resource name) of find item,
     *         or <p>null</p> if item not found
     */
    protected String getHref(String id, long principal)
            throws SyncSourceException {

        String href;

        try {
            href = im.getHref(id, principal, this.sourceURI);
        } catch (DataAccessException e) {
            throw new SyncSourceException("Error getting href:" +
                    e.getMessage(), e);
        }

        return href;

    }

    /**
     * Add item to new items array list
     *
     * @param id
     * @param crc
     * @param lastUpdate
     */
    protected void toSyncNewItems(String id, long crc, Date lastUpdate) {
        syncNewItems.add(new Item(id, lastUpdate, crc));
    }

    /**
     * Add item to update items array list
     *
     * @param id
     * @param crc
     * @param lastUpdate
     */
    protected void toSyncUpdateItems(String id, long crc, Date lastUpdate) {
        syncUpdateItems.add(new Item(id, lastUpdate, crc));
    }

    /**
     * Add item to remove items array list
     *
     * @param id
     */
    protected void toSyncRemoveItems(String id) {
        syncRemoveItems.add(new Item(id, null, 0));
    }

    /**
     * Add item to error new items array list
     *
     * @param id
     */
    protected void toSyncErrorNewItems(String id) {
        syncErrorNewItems.add(id);
    }

    /**
     * Add item to error update items array list
     *
     * @param id
     */
    protected void toSyncErrorUpdateItems(String id) {
        syncErrorUpdateItems.add(id);
    }

    /**
     * Add item to error remove items array list
     *
     * @param id
     */
    protected void toSyncErrorRemoveItems(String id) {
        syncErrorRemoveItems.add(id);
    }

    /**
     * Add SyncItem array to cacheSyncItems arrayList
     *
     * @param syncItems
     */
    protected void addToCacheSyncItems(SyncItem[] syncItems) {

        if (syncItems != null) {
            for (int i = 0, l = syncItems.length; i < l; i++) {
                cacheSyncItems.add(syncItems[i]);
            }
        }
    }

    /**
     * Add SyncItemKey with id to exchangeItemsKeys arrayList
     *
     * @param id
     */
    protected void addInExchangeItemsKeys(String id) {
        exchangeItemsKeys.add(new SyncItemKey(id));
    }

    /**
     * Remove SyncItemKey with id from exchangeItemsKeys arrayList
     *
     * @param id
     */
    protected void removeInExchangeItemsKeys(String id) {
        exchangeItemsKeys.remove(new SyncItemKey(id));
    }

    /**
     * find SyncItem in cacheSyncItems by id
     *
     * @param id
     */
    protected SyncItem getSyncItemById(String id) {

        SyncItem syncItem;

        if (cacheSyncItems != null) {

            Iterator i = cacheSyncItems.iterator();

            while (i.hasNext()) {
                syncItem = (SyncItem) i.next();
                if (id.equals(syncItem.getKey().getKeyAsString())) {
                    return syncItem;
                }

            }

        }

        return null;

    }

    /**
     * find SyncItemKey array about cacheSyncItemsKeys
     */
    protected SyncItemKey[] getCacheSyncItemsKeys() {

        int count = 0;

        SyncItemKey[] syncItemKeys = null;

        if (cacheSyncItems != null) {

            syncItemKeys = new SyncItemKey[cacheSyncItems.size()];

            Iterator i = cacheSyncItems.iterator();

            while (i.hasNext()) {
                syncItemKeys[count++] = new SyncItemKey(((SyncItem) i.next()).getKey().getKeyAsString());
            }
        }

        return syncItemKeys;

    }

    /**
     * find SyncItemKey array about items
     *
     * @param items
     */
    protected SyncItemKey[] getSyncItemKeysFromItems(Item[] items) {

        SyncItemKey[] syncItemKeys = null;

        if (items != null) {

            syncItemKeys = new SyncItemKey[items.length];

            for (int i = 0, l = items.length; i < l; i++) {
                syncItemKeys[i] = new SyncItemKey(items[i].getId());
            }
        }

        return syncItemKeys;

    }

    /**
     * return SyncItemKey ArrayList array about items
     *
     * @param items
     */
    protected ArrayList getSyncItemKeysArrayListFromItems(Item[] items) {

        ArrayList syncItemKeys;

        syncItemKeys = new ArrayList();

        if (items != null) {

            for (int i = 0, l = items.length; i < l; i++) {
                syncItemKeys.add(new SyncItemKey(items[i].getId()));
            }
        }

        return syncItemKeys;

    }

    /**
     * <p>Gets the list of cached items to be used in the package's classes and the subclasses.</p>
     *
     * @return the value of cacheSyncItems
     */
    protected List getCachedSyncItems() {
        return this.cacheSyncItems;
    }

    /**
     * get Item by id
     *
     * @param items array
     * @param id
     * @return <p>item</p> if find
     *         <p>null</p> if not find
     */
    private Item getItemById(Item[] items,
                             String id) {

        Item item = null;
        Item sequenceItem;

        for (int i = 0, l = items.length; i < l; i++) {

            sequenceItem = items[i];

            if (sequenceItem.getId().equals(id)) {
                item = sequenceItem;
                break;
            }

        }

        return item;

    }

    /**
     * check if item id is included
     * in ids from array of items
     *
     * @param items items array
     * @param id    id to find
     * @return <p>true</p> if is included
     *         <p>false</p> if is not included
     */
    private boolean isIdInArray(Item[] items,
                                String id) {

        boolean isFind = false;

        for (int i = 0, l = items.length; i < l; i++) {

            if (items[i].getId().equals(id)) {
                isFind = true;
                break;
            }

        }

        return isFind;

    }

    /**
     * check if item is included
     * in array of items
     *
     * @param items items array
     * @param item  item to find
     * @return <p>true</p> if is included
     *         <p>false</p> if is not included
     */
    private boolean isItemInArray(Item[] items,
                                  Item item) {

        return isIdInArray(items, item.getId());

    }
}