/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.engine.source;

import com.funambol.exchange.DataAccessException;
import com.funambol.exchange.items.calendar.manager.CalendarManager;
import com.funambol.exchange.items.calendar.model.Calendar;
import com.funambol.exchange.items.common.manager.ItemManager;
import com.funambol.exchange.items.common.model.Item;
import com.funambol.exchange.util.Def;
import com.funambol.exchange.util.StringTools;
import com.funambol.foundation.common.merge.MergeResult;
import com.funambol.foundation.pdi.converter.CalendarToIcalendar;
import com.funambol.foundation.pdi.converter.CalendarToXML;
import com.funambol.foundation.pdi.converter.ConverterException;
import com.funambol.foundation.pdi.parser.ICalendarParser;
import com.funambol.foundation.pdi.parser.XMLEventParser;
import com.funambol.foundation.pdi.utils.SourceUtils;
import com.funambol.framework.core.AlertCode;
import com.funambol.framework.engine.SyncItem;
import com.funambol.framework.engine.SyncItemImpl;
import com.funambol.framework.engine.SyncItemKey;
import com.funambol.framework.engine.source.MergeableSyncSource;
import com.funambol.framework.engine.source.SyncSourceException;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.server.store.NotFoundException;

import java.io.ByteArrayInputStream;
import java.io.Serializable;
import java.sql.Timestamp;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;


/**
 * This class define <i>SyncSource</i>
 * between SyncServer and Exchange Server calendar items
 *
 * @version $Id: ExchangeCalendarSyncSource.java,v 1.1 2006/08/08 07:52:15 gibi Exp $
 */
public class ExchangeCalendarSyncSource extends ExchangeSyncSource
        implements MergeableSyncSource, Serializable {

    // --------------------------------------------------------------- Constants

    private static final String EXCHANGE_HREF_EXTENSION = ".eml";

    // ------------------------------------------------------------ Private data

    private CalendarManager cm = null;

    // ------------------------------------------------------------ Constructors

    public ExchangeCalendarSyncSource() {
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * init manager
     */
    public void initManager()
            throws SyncSourceException {

        try {

            this.cm = new CalendarManager(this.host,
                    this.port,
                    this.exchangeHost,
                    this.deviceTimeZone);

        } catch (Exception e) {

            throw new SyncSourceException("Error init manager: " +
                    e.getMessage(), e);
        }

    }

    /**
     * @see com.funambol.framework.engine.source.SyncSource
     */
    public SyncItem addSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        Calendar calendar;
        Calendar newCalendar = null;
        String href;
        String itemKey;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("addSyncItem - source: " + sourceURI);
        }

        try {

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // created
            // default href = this.CurrentTimesMillis()
            //
            href = String.valueOf(System.currentTimeMillis()) +
                    EXCHANGE_HREF_EXTENSION;

            if (log.isLoggable(Level.FINEST)) {
                log.finest(itemKey + " is new: adding");
            }

            calendar = getCalendarFromSyncItem(syncItem);

            calendar.setHref(href);

            newCalendar = cm.setCalendar(calendar,
                    username,
                    credentials,
                    exchangeFolder);

            long crc = StringTools.getCRCFromString(newCalendar.getDtStart().getPropertyValueAsString() +
                    newCalendar.getDtEnd().getPropertyValueAsString() +
                    newCalendar.getLocation().getPropertyValueAsString());

            toSyncNewItems(newCalendar.getId(), crc, newCalendar.getLastUpdate());
            addInExchangeItemsKeys(newCalendar.getId());

            syncItem = getSyncItem(newCalendar);

            return syncItem;

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "addSyncItem", e);
            }

            if (newCalendar != null) {
                toSyncErrorNewItems(newCalendar.getId());
            }

            throw new SyncSourceException("Error add the item " +
                    syncItem +
                    " " +
                    e.getMessage(), e);
        }

    }

    /**
     * @see com.funambol.framework.engine.source.SyncSource
     */
    public SyncItem updateSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        Calendar clientCalendar;
        Calendar updatedCalendar = null;

        String href;
        String itemKey;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("updateSyncItem - source: " + sourceURI);
        }

        try {

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // create href [eg. 1234.eml] in exchange items
            //
            href = getHref(itemKey, principalId);

            if (log.isLoggable(Level.FINEST)) {
                log.finest(itemKey + " already exists: updating");
            }

            clientCalendar = getCalendarFromSyncItem(syncItem);

            if (log.isLoggable(Level.INFO)) {
                log.info("Merging server and client item to ensure data is not lost.");
            }

            clientCalendar.setHref(href);

            updatedCalendar = cm.setCalendar(clientCalendar,
                    username,
                    credentials,
                    exchangeFolder);

            long crc = StringTools.getCRCFromString(updatedCalendar.getDtStart().getPropertyValueAsString() +
                    updatedCalendar.getDtEnd().getPropertyValueAsString() +
                    updatedCalendar.getLocation().getPropertyValueAsString());

            toSyncUpdateItems(updatedCalendar.getId(), crc, updatedCalendar.getLastUpdate());

            syncItem = getSyncItem(updatedCalendar);

            return syncItem;

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "updateSyncItem", e);
            }

            if (updatedCalendar != null) {
                toSyncErrorUpdateItems(updatedCalendar.getId());
            }

            throw new SyncSourceException("Error setting the item " +
                    syncItem + " " + e.getMessage(), e);
        }

    }

    /**
     * @see com.funambol.framework.engine.source.SyncSource
     */
    public SyncItem getSyncItemFromId(SyncItemKey syncItemKey)
            throws SyncSourceException {

        Calendar calendar = null;
        String id;
        SyncItem syncItem;

        id = syncItemKey.getKeyAsString();

        //
        // search syncItem in cache
        //
        syncItem = getSyncItemById(id);

        if (syncItem != null) {
            return syncItem;
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe
                    ("Calendar with id " + id + " not found while reading Exchange cache");
        }

        try {

            calendar = cm.getCalendarById(username,
                    credentials,
                    id,
                    this.exchangeFolder);
        } catch (Exception e) {

            Throwable previous = e.getCause();

            if (previous instanceof NotFoundException) {

                if (log.isLoggable(Level.SEVERE)) {
                    log.severe
                            ("Calendar with " + id + " not found while reading Exchange database: " +
                                    e.getMessage());
                }

                Sync4jLogger.getLogger(Def.LOGGER_NAME).
                        throwing(getClass().
                                getName(), "readExchangeDatabase", e);

            } else {

                if (log.isLoggable(Level.FINEST)) {
                    log.severe("SyncSource error: " + e.getMessage());
                    log.throwing(getClass().getName(), "getSyncItemFromId", e);
                }

                throw new SyncSourceException("Error reading items", e);
            }
        }

        if (calendar != null) {
            calendar.setModificationStatus(getSyncItemStatus(calendar.getId()));
            syncItem = getSyncItem(calendar);
        }

        return syncItem;
    }

    /**
     * @see com.funambol.framework.engine.source.SyncSource
     */
    public void removeSyncItem(SyncItemKey itemKey, Timestamp time, boolean softDelete)
            throws SyncSourceException {

        if (!softDelete) {
            removeSyncItem(itemKey);
        }

    }

    /**
     * @see com.funambol.framework.engine.source.SyncSource
     */
    public SyncItemKey[] getSyncItemKeysFromTwin(SyncItem syncItem)
            throws SyncSourceException {

        Item[] items;

        try {

            Calendar calendar = getCalendarFromSyncItem(syncItem);
            items = this.cm.getTwinItems(calendar,
                    principalId,
                    this.sourceURI);


        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "getSyncItemKeysFromTwin", e);
            }

            throw new SyncSourceException("Error getting twin", e);
        }

        return getSyncItemKeysFromItems(items);
    }

    /**
     * Merges server item and client item.
     *
     * @param serverKey  the item's key on the server
     * @param clientItem the item on the client
     * @return true is the content of the client item is changed
     * @throws SyncSourceException
     */
    public boolean mergeSyncItems(SyncItemKey serverKey, SyncItem clientItem)
            throws SyncSourceException {

        //return false;

        Calendar clientCalendar;
        Calendar serverCalendar;
        com.funambol.foundation.pdi.event.Calendar foundationCalendar;
        String content;

        Timestamp t = clientItem.getTimestamp();

        try {


            byte[] itemContent = clientItem.getContent();

            if (itemContent == null) {
                itemContent = new byte[0];
            }

            content = new String(itemContent);

            //
            // 1. get the Calendar object from the itemA (this is the item sent by the client)
            //
            clientCalendar = getCalendar(clientItem.getKey().getKeyAsString(), content, t);

            //
            // 3. Read the server calendar
            //
            try {
                serverCalendar = getCalendarFromSyncItem(getSyncItemFromId(serverKey));
            } catch (Exception ex) {
                throw new SyncSourceException("Error reading the calendar '" +
                        serverKey.getKeyAsString() + "'", ex);
            }

            MergeResult mergeResult;

            try {
                mergeResult = clientCalendar.merge(serverCalendar);
            } catch (Exception ex) {
                throw new SyncSourceException("Error merging the calendars '" +
                        serverKey.getKeyAsString() + "'", ex);
            }

            if (mergeResult.isSetBRequired()) {
                //
                // We have to update the server calendar
                //
                SyncItem serverSyncItem = getSyncItem(serverCalendar);
                serverSyncItem.setTimestamp(t);
                updateSyncItem(serverSyncItem);
                updateCalendarInCache(serverKey.getKeyAsString(),
                        serverCalendar, t);

            }

            if (mergeResult.isSetARequired()) {
                //converting the Calendar object into a vcard string
                foundationCalendar = new com.funambol.foundation.pdi.event.Calendar();
                foundationCalendar.setEvent(serverCalendar);
                if (TYPE_ICAL.equals(getType())) {
                    content = calendar2ical(foundationCalendar);
                } else {
                    content = calendar2sife(foundationCalendar);
                }
                clientItem.setContent(content.getBytes());
                clientItem.setTimestamp(t);
                SyncItem clientSyncItem = getSyncItem(clientCalendar);
                clientSyncItem.setTimestamp(t);
            }

            if (log.isLoggable(Level.FINEST)) {
                log.finest("MergeResult: " + mergeResult);
            }

            return mergeResult.isSetARequired();

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "mergeSyncItems", e);
            }

            throw new SyncSourceException("Error merging the calendars", e);
        }

    }

    /**
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @return items array
     */
    public Item[] getExchangeItems(String username,
                                   String credentials,
                                   String exchangeFolder,
                                   String[] ids,
                                   boolean isBeginSync)
            throws SyncSourceException {

        Item     [] items;
        Calendar [] calendars;

        try {

            //
            // is not slow sync or called from endSync
            // load calendar with only fields for twin setting
            //
            if (((syncMode != AlertCode.SLOW) &&
                    (syncMode != AlertCode.REFRESH_FROM_SERVER)) ||
                    !isBeginSync) {

                calendars = this.cm.getCalendarsForTwin(username,
                        credentials,
                        ids,
                        exchangeFolder);


            } else {

                calendars = this.cm.getCalendars(username,
                        credentials,
                        ids,
                        exchangeFolder);

                addToCacheSyncItems(getSyncItems(calendars));

            }

            items = getItems(calendars);


        } catch (DataAccessException e) {
            throw new SyncSourceException("Error getting items", e);
        }

        return items;

    }

    /**
     * @param username
     * @param credentials
     * @param ids
     * @param exchangeFolder
     * @return items array
     */
    public SyncItem[] getSyncItems(String username,
                                   String credentials,
                                   String[] ids,
                                   String exchangeFolder)
            throws SyncSourceException {

        SyncItem [] syncItems;
        Calendar [] calendars;

        try {

            calendars = this.cm.getCalendars(username,
                    credentials,
                    ids,
                    exchangeFolder);

            syncItems = getSyncItems(calendars);


        } catch (DataAccessException e) {
            throw new SyncSourceException("Error getting items", e);
        }

        return syncItems;

    }

    // --------------------------------------------------------- Private methods

    private Calendar getCalendar(String id,
                                 String content,
                                 Date t)
            throws SyncSourceException {

        Calendar c = new Calendar(id);

        c.setLastUpdate(t);

        if (content != null && content.length() > 0) {

            if (TYPE_ICAL.equals(getType())) {
                ical2Calendar(content, c);
            } else {
                sife2Calendar(content, c);
            }
        }

        return c;

    }

    /**
     * Create SyncItem array from Calendar array
     *
     * @param exchangeCalendars
     * @return array of SyncItem
     * @throws SyncSourceException
     */
    private SyncItem[] getSyncItems(Calendar[] exchangeCalendars)
            throws SyncSourceException {

        SyncItem[] syncItems;

        syncItems = new SyncItem[exchangeCalendars.length];

        for (int i = 0, l = syncItems.length; i < l; i++) {
            syncItems[i] = getSyncItem(exchangeCalendars[i]);
        }

        return syncItems;

    }

    /**
     * Create SyncItem from Calendar
     *
     * @param exchangeCalendar
     * @return SyncItem
     * @throws SyncSourceException
     */
    private SyncItem getSyncItem(Calendar exchangeCalendar)
            throws SyncSourceException {

        SyncItem syncItem;
        String content;

        com.funambol.foundation.pdi.event.Calendar
                calendar = new com.funambol.foundation.pdi.event.Calendar();

        syncItem = new SyncItemImpl(this,
                exchangeCalendar.getId(),
                exchangeCalendar.getModificationStatus());

        calendar.setEvent(exchangeCalendar);
        if (TYPE_ICAL.equals(getType())) {
            content = calendar2ical(calendar);
        } else {
            content = calendar2sife(calendar);
        }

        syncItem.setContent(content.getBytes());

        syncItem.setType(this.getType());


        return syncItem;

    }

    /**
     * Delete all calendars from the Exchange server
     *
     * @see ExchangeSyncSource
     */
    public void deleteAll()
            throws SyncSourceException {
        if (log.isLoggable(Level.INFO)) {
            log.log(Level.INFO, "Removing all calendars from Exchange for user " + this.username + ".");
        }

        Calendar[] allExcCalendars;
        try {
            allExcCalendars = cm.getCalendars(this.username, this.credentials, new String[0], this.getExchangeFolder());
        }
        catch (DataAccessException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.log(Level.SEVERE, "Error removing all calendars from Exchange.");
            }
            throw new SyncSourceException(e);
        }

        for (int i = 0; i < allExcCalendars.length; i++) {
            try {
                cm.removeCalendar(allExcCalendars[i], this.username, this.credentials, this.exchangeFolder);
            }
            catch (DataAccessException e) {
                if (log.isLoggable(Level.SEVERE)) {
                    log.log(Level.SEVERE, "Error removing calendar with id " +
                            allExcCalendars[i].getId() + " from Exchange.");
                }
                if (log.isLoggable(Level.INFO)) {
                    log.log(Level.INFO, e.getMessage(), e);
                }
            }
        }
        if (log.isLoggable(Level.INFO)) {
            log.log(Level.INFO, "Clearing cached events for source uri " +
                    this.sourceURI + " and principal " + this.principalId);
        }
        try {
            new ItemManager().deleteItems(cm.getAllItems(this.sourceURI, this.principalId),
                    this.sourceURI, this.principalId);
        }
        catch (DataAccessException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.log(Level.SEVERE, "Error while clearing cached events: " + e.getMessage() + ".");
            }
            if (log.isLoggable(Level.INFO)) {
                log.log(Level.INFO, e.getMessage(), e);
            }
        }
    }

    /**
     * Create Item array from Calendar array
     *
     * @param exchangeCalendars
     * @return array of SyncItem
     */
    private Item[] getItems(Calendar[] exchangeCalendars) {

        Item[] items;

        items = new Item[exchangeCalendars.length];

        for (int i = 0, l = items.length; i < l; i++) {
            items[i] = getItem(exchangeCalendars[i]);
        }

        return items;

    }

    /**
     * Create Item from Calendar
     *
     * @param exchangeCalendar
     * @return item
     */
    private Item getItem(Calendar exchangeCalendar) {

        Item item;

        item = new Item(exchangeCalendar.getId(),
                exchangeCalendar.getHref(),
                exchangeCalendar.getLastUpdate(),
                StringTools.getCRCFromString(
                        exchangeCalendar.getDtStart().getPropertyValueAsString() +
                                exchangeCalendar.getDtEnd().getPropertyValueAsString() +
                                exchangeCalendar.getLocation().getPropertyValueAsString()));
        return item;

    }

    ///**
    // * Create SyncItemKey array from Calendar array
    // *
    // * @param calendars
    // * @return SyncItemKey array from Calendar array
    // */
    //private SyncItemKey[] getSyncItemKeysFromCalendars(Calendar[] calendars) {
    //
    //   SyncItemKey[] syncItemKeys;
    //
    //   if (calendars == null) {
    //       return null;
    //   }
    //
    //   syncItemKeys = new SyncItemKey[calendars.length];
    //
    //   for (int i = 0, l = calendars.length; i < l; i++) {
    //      syncItemKeys[i] = new SyncItemKey(calendars[i].getId());
    //   }
    //
    //   return syncItemKeys;
    //
    //}

    /**
     * Converts the given calendar into a ical String
     *
     * @param c the calendar to convert
     * @return the ical
     * @throws SyncSourceException in case of convertion errors
     */
    private String calendar2ical(com.funambol.foundation.pdi.event.Calendar c)
            throws SyncSourceException {
        try {
            return new CalendarToIcalendar(deviceTimeZone, deviceCharset).convert(c);
        } catch (ConverterException e) {
            throw new SyncSourceException("Conversion error for item "
                    + ((Calendar) c.getEvent()).getId()
                    + ": "
                    + e.getMessage()
                    , e
            );
        }
    }

    /**
     * Converts the given calendar into a sife String
     *
     * @param c the calendar to convert
     * @return the sife document
     * @throws SyncSourceException in case of convertion errors
     */
    private String calendar2sife(com.funambol.foundation.pdi.event.Calendar c)
            throws SyncSourceException {
        try {
            return new CalendarToXML(deviceTimeZone, deviceCharset).convert(c);
        } catch (ConverterException e) {
            throw new SyncSourceException("Convertion error for item "
                    + ((Calendar) c.getEvent()).getId()
                    + ": "
                    + e.getMessage()
                    , e
            );
        }
    }

    /**
     * Fill the given CalEvent object with the properties in the given ical
     *
     * @param content iCal content
     * @param c       the calendar to feed
     * @throws SyncSourceException
     */
    private void ical2Calendar(String content, Calendar c)
            throws SyncSourceException {

        //
        // just make sure we have something to do...
        //
        if ((content == null) || (content.length() == 0)) {
            return;
        }


        ByteArrayInputStream is = null;
        ICalendarParser parser;

        com.funambol.foundation.pdi.event.Calendar calendar;
        com.funambol.foundation.pdi.event.Event event;

        try {

            /**
             * This is used to replace the String "=\r\n" with the
             * String "\r\n ": This replace is necessary because some devices do
             * not send the correct line delimiter for long lines.
             */
            //content = content.replaceAll("=\r\n","\r\n ");
            content = SourceUtils.handleLineDelimiting(content);


            is = new ByteArrayInputStream(content.getBytes());

            //converting the vcard into a Calendar object

            parser = new ICalendarParser(is, deviceTimeZoneDescr, deviceCharset);
            parser.setLogger(log);

            calendar = parser.iCalendar();

            //
            // We now have to convert a portal calendar into a foundation calendar
            //
            event = calendar.getEvent();

            c.setReminder(event.getReminder());
            c.setCategories(event.getCategories());
            c.setClassEvent(event.getClassEvent());
            c.setDescription(event.getDescription());
            c.setLatitude(event.getLatitude());
            c.setLongitude(event.getLongitude());
            c.setLocation(event.getLocation());
            c.setPriority(event.getPriority());
            c.setStatus(event.getStatus());
            c.setSummary(event.getSummary());
            c.setDtEnd(event.getDtEnd());
            c.setDtStart(event.getDtStart());
            c.setDuration(event.getDuration());
            c.setTransp(event.getTransp());
            c.setOrganizer(event.getOrganizer());
            c.setUrl(event.getUrl());
            c.setUid(event.getUid());
            c.setContact(event.getContact());
            c.setCreated(event.getCreated());
            c.setDtStamp(event.getDtStamp());
            c.setLastModified(event.getLastModified());
            c.setSequence(event.getSequence());
            c.setXTag(event.getXTags());
            c.setAllDay(Boolean.valueOf(event.isAllDay()));
            c.setMeetingStatus(event.getMeetingStatus());
            c.setMileage(event.getMileage());
            c.setReplyTime(event.getReplyTime());
            c.setRecurrencePattern(event.getRecurrencePattern());

        } catch (Exception e) {
            throw new SyncSourceException("Convertion error for item "
                    + c.getId()
                    + ": "
                    + e.getMessage()
                    , e
            );
        } finally {
            if (is != null) {
                try {
                    is.close();
                }
                catch (Exception e) {
                    if (log.isLoggable(Level.ALL)) {
                        log.log(Level.ALL, e.getMessage());
                    }
                }
            }

        }
    }

    /**
     * Fill the given CalEvent object with the properties in the given sife
     * document
     *
     * @param content  iCal content
     * @param calendar the calendar to feed
     * @throws SyncSourceException
     */
    private void sife2Calendar(String content, Calendar calendar)
            throws SyncSourceException {
        //
        // just make sure we have something to do...
        //
        if ((content == null) || (content.length() == 0)) {
            return;
        }


        ByteArrayInputStream is = null;
        XMLEventParser parser;
        com.funambol.foundation.pdi.event.Calendar foundationCalendar;

        foundationCalendar = new com.funambol.foundation.pdi.event.Calendar();

        try {
            is = new ByteArrayInputStream(content.getBytes());

            //converting the vcard into a Calendar object

            parser = new XMLEventParser(is);

            parser.parse(foundationCalendar);

            calendar.copy(foundationCalendar.getEvent());

        } catch (Exception e) {
            throw new SyncSourceException("Parsing error for item "
                    + calendar.getId()
                    + ": "
                    + e.getMessage()
                    , e
            );
        } finally {
            if (is != null) {
                try {
                    is.close();
                }
                catch (Exception e) {
                    if (log.isLoggable(Level.ALL)) {
                        log.log(Level.ALL, e.getMessage());
                    }
                }
            }
        }
    }

    /**
     * Creates a new Calendar from the given syncItem
     *
     * @param syncItem SyncItem
     * @return Calendar
     */
    private Calendar getCalendarFromSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        String content;
        Calendar calendar;

        String itemKey = syncItem.getKey().getKeyAsString();

        byte[] itemContent =
                syncItem.getContent();

        if (itemContent == null) {
            itemContent = new byte[0];
        }

        content = new String(itemContent);

        Date t = syncItem.getTimestamp();

        calendar = getCalendar(itemKey, content, t);

        return calendar;
    }

    /**
     *
     */
    private void removeSyncItem(SyncItemKey itemKey)
            throws SyncSourceException {

        Calendar calendar;

        String id;
        String href;

        id = itemKey.getKeyAsString();

        href = getHref(id, principalId);

        try {
            if (href != null) {
                calendar = new Calendar(id);
                calendar.setHref(href);

                this.cm.removeCalendar(calendar, username, credentials, exchangeFolder);
                toSyncRemoveItems(calendar.getId());
                removeInExchangeItemsKeys(calendar.getId());
            } else {
                if (log.isLoggable(Level.SEVERE)) {
                    log.severe
                            ("Remove calendar failed. Calendar with " + id + " not found while reading fnbl_exch_cache");
                }
                toSyncErrorRemoveItems(id);
            }
        } catch (Exception e) {

            toSyncErrorRemoveItems(itemKey.getKeyAsString());

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "removeSyncItem", e);
            }

            throw new SyncSourceException("Error removing items", e);
        }
    }

    /**
     * <p>Updates the SyncItem with the the input key from the SyncSource's cache
     *  with the values from the input Event and sent the input timezone for it.</p>
     * @param chachedItemKey (in) key of the Sync Item in the SyncSource's cache
     * @param eventToSet (in) Event to set in the sync item from cache
     * @param timeToSet the timezone to set for the sync item from cache
     * @throws SyncSourceException is an error occurs
     */
    private void updateCalendarInCache(String chachedItemKey,
                                       com.funambol.foundation.pdi.event.Event eventToSet,
                                       Timestamp timeToSet)
            throws SyncSourceException {
        List cachedSyncItems;
        Iterator iterator;
        SyncItem syncItem;
        String content, key;
        com.funambol.foundation.pdi.event.Calendar foundationCalendar;


        cachedSyncItems = this.getCachedSyncItems();

        iterator = cachedSyncItems.iterator();

        while (iterator.hasNext()) {
            syncItem = (SyncItem) iterator.next();
            if (syncItem == null || syncItem.getKey() == null) {
                continue;
            }

            key = syncItem.getKey().getKeyAsString();

            if (key.equals(chachedItemKey)) {
                foundationCalendar = new com.funambol.foundation.pdi.event.Calendar();
                foundationCalendar.setEvent(eventToSet);
                if (TYPE_ICAL.equals(getType())) {

                    content = calendar2ical(foundationCalendar);
                } else {
                    content = calendar2sife(foundationCalendar);
                }

                syncItem.setContent(content.getBytes());
                syncItem.setTimestamp(timeToSet);
                break;
            }
        }
    }
}
