/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.engine.source;

import com.funambol.exchange.DataAccessException;
import com.funambol.exchange.items.common.manager.ItemManager;
import com.funambol.exchange.items.common.model.Item;
import com.funambol.exchange.items.contact.manager.ContactManager;
import com.funambol.exchange.items.contact.model.Contact;
import com.funambol.exchange.util.Def;
import com.funambol.exchange.util.StringTools;
import com.funambol.foundation.common.merge.MergeResult;
import com.funambol.foundation.pdi.contact.Name;
import com.funambol.foundation.pdi.converter.ContactToVcard;
import com.funambol.foundation.pdi.converter.ContactToXML;
import com.funambol.foundation.pdi.converter.ConverterException;
import com.funambol.foundation.pdi.parser.VcardParser;
import com.funambol.foundation.pdi.parser.XMLContactParser;
import com.funambol.foundation.pdi.utils.SourceUtils;
import com.funambol.framework.core.AlertCode;
import com.funambol.framework.engine.SyncItem;
import com.funambol.framework.engine.SyncItemImpl;
import com.funambol.framework.engine.SyncItemKey;
import com.funambol.framework.engine.source.MergeableSyncSource;
import com.funambol.framework.engine.source.SyncSourceException;
import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.server.store.NotFoundException;

import java.io.ByteArrayInputStream;
import java.io.Serializable;
import java.sql.Timestamp;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;


/**
 * This class define <i>SyncSource</i>
 * between SyncServer and Exchange Server contacts items
 *
 * @version $Id: ExchangeContactSyncSource.java,v 1.1 2006/08/08 07:52:15 gibi Exp $
 */
public class ExchangeContactSyncSource extends ExchangeSyncSource
        implements MergeableSyncSource, Serializable {

    // --------------------------------------------------------------- Constants

    private static final String EXCHANGE_HREF_EXTENSION = ".eml";

    // ------------------------------------------------------------ Private data

    private ContactManager cm = null;

    // ------------------------------------------------------------ Constructors

    public ExchangeContactSyncSource() {
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * init manager
     */
    public void initManager()
            throws SyncSourceException {

        try {

            this.cm = new ContactManager(this.host,
                    this.port,
                    this.exchangeHost,
                    this.deviceTimeZone);

        } catch (Exception e) {

            throw new SyncSourceException("Error init manager: " +
                    e.getMessage(), e);
        }

    }

    /**
     * @see com.funambol.framework.engine.source.SyncSource
     */
    public SyncItem addSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        Contact contact;
        Contact newContact = null;

        String itemKey;
        String href;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("addSyncItem - source: " + sourceURI);
        }

        try {

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // created
            // default href = this.CurrentTimesMillis()
            //
            href = String.valueOf(System.currentTimeMillis()) +
                    EXCHANGE_HREF_EXTENSION;

            if (log.isLoggable(Level.FINEST)) {
                log.finest(itemKey + " is new: adding");
            }

            contact = getContactFromSyncItem(syncItem);
            contact.setHref(href);

            newContact = cm.setContact(contact,
                    username,
                    credentials,
                    exchangeFolder);

            String firstName = null;
            String lastName = null;

            Name name = newContact.getName();

            if (name != null) {
                if (name.getFirstName() != null) {
                    firstName = name.getFirstName().
                            getPropertyValueAsString();
                }
                if (name.getLastName() != null) {
                    lastName = name.getLastName().
                            getPropertyValueAsString();
                }
            }

            long crc = StringTools.getCRCFromString(firstName + lastName);

            toSyncNewItems(newContact.getId(), crc, newContact.getLastUpdate());
            addInExchangeItemsKeys(newContact.getId());

            syncItem = getSyncItem(newContact);

            return syncItem;

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "addSyncItem", e);
            }

            if (newContact != null) {
                toSyncErrorNewItems(newContact.getId());
            }

            throw new SyncSourceException("Error add the item "
                    + syncItem
                    , e
            );
        }
    }

    /**
     * @see com.funambol.framework.engine.source.SyncSource
     */
    public SyncItem updateSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        Contact clientContact;
        Contact updateContact = null;

        String itemKey;
        String href;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("updateSyncItem - source: " + sourceURI);
        }

        try {

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // create href [eg. 1234.eml] in exchange items
            //
            href = getHref(itemKey, principalId);

            if (log.isLoggable(Level.FINEST)) {
                log.finest(itemKey + " already exists: updating");
            }

            clientContact = getContactFromSyncItem(syncItem);

            clientContact.setHref(href);

            updateContact = cm.setContact(clientContact,
                    username,
                    credentials,
                    exchangeFolder);

            String firstName = null;
            String lastName = null;

            Name name = updateContact.getName();

            if (name != null) {
                if (name.getFirstName() != null) {
                    firstName = name.getFirstName().
                            getPropertyValueAsString();
                }
                if (name.getLastName() != null) {
                    lastName = name.getLastName().
                            getPropertyValueAsString();
                }
            }

            long crc = StringTools.getCRCFromString(firstName + lastName);

            toSyncNewItems(updateContact.getId(), crc, updateContact.getLastUpdate());

            syncItem = getSyncItem(updateContact);

            return syncItem;

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "updateSyncItem", e);
            }

            if (updateContact != null) {
                toSyncErrorUpdateItems(updateContact.getId());
            }

            throw new SyncSourceException("Error setting the item " +
                    syncItem, e);
        }
    }

    /**
     * @see com.funambol.framework.engine.source.SyncSource
     */
    public SyncItem getSyncItemFromId(SyncItemKey syncItemKey)
            throws SyncSourceException {

        Contact contact = null;
        String id;
        SyncItem syncItem;

        id = syncItemKey.getKeyAsString();

        //
        // search syncItem in cache
        //
        syncItem = getSyncItemById(id);

        if (syncItem != null) {
            return syncItem;
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe
                    ("Contact with id " + id + " not found while reading Exchange cache");
        }

        try {

            contact = cm.getContactById(username,
                    credentials,
                    id,
                    exchangeFolder);

        } catch (Exception e) {

            Throwable previous = e.getCause();

            if (previous instanceof NotFoundException) {

                if (log.isLoggable(Level.SEVERE)) {
                    log.severe
                            ("Contact not found while reading Exchange database: " +
                                    e.getMessage());
                }

                Sync4jLogger.getLogger(Def.LOGGER_NAME).
                        throwing(getClass().
                                getName(), "readExchangeDatabase", e);

            } else {

                if (log.isLoggable(Level.FINEST)) {
                    log.severe("SyncSource error: " + e.getMessage());
                    log.throwing(getClass().getName(), "getSyncItemFromId", e);
                }

                throw new SyncSourceException("Error reading items", e);
            }
        }

        if (contact != null) {
            contact.setStatus(getSyncItemStatus(contact.getId()));
            syncItem = getSyncItem(contact);
        }

        return syncItem;
    }

    /**
     * @see com.funambol.framework.engine.source.SyncSource
     */
    public void removeSyncItem(SyncItemKey itemKey, Timestamp time, boolean softDelete)
            throws SyncSourceException {

        if (!softDelete) {
            removeSyncItem(itemKey);
        }

    }

    /**
     * The same as <i>getSyncItemKeysFromTwin()</i>
     *
     * @see com.funambol.framework.engine.source.SyncSource
     */
    public SyncItemKey[] getSyncItemKeysFromTwin(SyncItem syncItem)
            throws SyncSourceException {

        Item[] items;

        try {

            Contact contact = getContactFromSyncItem(syncItem);
            items = this.cm.getTwinItems(contact,
                    principalId,
                    this.sourceURI);
        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "getSyncItemKeysFromTwin", e);
            }

            throw new SyncSourceException("Error getting twin", e);
        }

        return getSyncItemKeysFromItems(items);
    }

    /**
     * Merges server item and client item.
     *
     * @param serverKey  the item's key on the server
     * @param clientItem the item on the client
     * @return true is the content of the client item is changed
     * @throws SyncSourceException
     */
    public boolean mergeSyncItems(SyncItemKey serverKey, SyncItem clientItem)
            throws SyncSourceException {

        //return false;

        Contact clientContact;
        Contact serverContact;

        Timestamp t = clientItem.getTimestamp();

        try {

            byte[] itemContent = clientItem.getContent();

            if (itemContent == null) {
                itemContent = new byte[0];
            }

            String content = new String(itemContent);

            //
            // 1. get the Contact object from the itemA (this is the item sent by the client)
            //
            if (log.isLoggable(Level.INFO)) {
                log.info("Merging server item " + serverKey.getKeyAsString() + " with its client correspondent.");
            }
            clientContact = getContact(clientItem.getKey().getKeyAsString(), content, t);

            //
            // 3. Read the server contact
            //
            try {
                serverContact = getContactFromSyncItem(getSyncItemFromId(serverKey));
            } catch (Exception ex) {
                throw new SyncSourceException("Error reading the contact '" +
                        serverKey.getKeyAsString() + "'", ex);
            }

            MergeResult mergeResult;

            try {
                mergeResult = clientContact.merge(serverContact);
            } catch (Exception ex) {
                throw new SyncSourceException("Error merging the contacts '" +
                        serverKey.getKeyAsString() + "'", ex);
            }

            if (mergeResult.isSetBRequired()) {
                //
                // We have to update the server contact
                //
                SyncItem serverSyncItem = getSyncItem(serverContact);
                serverSyncItem.setTimestamp(t);
                updateSyncItem(serverSyncItem);
                updateContactInCache(serverSyncItem.getKey().getKeyAsString(),
                        serverContact, t);

            }


            if (mergeResult.isSetARequired()) {
                //converting the Contact object into a vcard string
                if (TYPE_VCARD.equals(getType())) {
                    content = contact2vcard(serverContact);
                } else {
                    content = contact2sifc(serverContact);
                }
                clientItem.setContent(content.getBytes());
                clientItem.setTimestamp(t);
            }

            if (log.isLoggable(Level.FINEST)) {
                log.finest("MergeResult: " + mergeResult);
            }

            return mergeResult.isSetARequired();

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "mergeSyncItems", e);
            }

            throw new SyncSourceException("Error merging contacts", e);
        }

    }

    /**
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @return items array
     */
    public Item[] getExchangeItems(String username,
                                   String credentials,
                                   String exchangeFolder,
                                   String[] ids,
                                   boolean isBeginSync)
            throws SyncSourceException {

        Item     [] items;
        Contact  [] contacts;

        try {

            //
            // is not slow sync or called from endSync
            // load calendar with only fields for twin setting
            //
            if (((syncMode != AlertCode.SLOW) &&
                    (syncMode != AlertCode.REFRESH_FROM_SERVER)) ||
                    !isBeginSync) {

                contacts = this.cm.getContactsForTwin(username,
                        credentials,
                        ids,
                        exchangeFolder);


            } else {

                contacts = this.cm.getContacts(username,
                        credentials,
                        ids,
                        exchangeFolder);

                addToCacheSyncItems(getSyncItems(contacts));

            }

            items = getItems(contacts);


        } catch (DataAccessException e) {
            throw new SyncSourceException("Error getting items", e);
        }

        return items;

    }

    /**
     * @param username
     * @param credentials
     * @param ids
     * @param exchangeFolder
     * @return SyncItem array
     */
    public SyncItem[] getSyncItems(String username,
                                   String credentials,
                                   String[] ids,
                                   String exchangeFolder)
            throws SyncSourceException {

        SyncItem [] syncItems;
        Contact  [] contacts;

        try {

            contacts = this.cm.getContacts(username,
                    credentials,
                    ids,
                    exchangeFolder);

            syncItems = getSyncItems(contacts);


        } catch (DataAccessException e) {
            throw new SyncSourceException("Error getting items", e);
        }

        return syncItems;

    }

    /**
     * <p>Delete all contacts from the Exchange server.</p>
     *
     * @see ExchangeSyncSource
     */
    public void deleteAll()
            throws SyncSourceException {
        if (log.isLoggable(Level.INFO)) {
            log.log(Level.INFO, "Removing all contacts from Exchange for user " + this.username + ".");
        }

        Contact[] allExcContacts;
        try {
            allExcContacts = cm.getContacts(this.username, this.credentials, new String[0], this.getExchangeFolder());
        }
        catch (DataAccessException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.log(Level.SEVERE, "Error removing all contacts from Exchange.");
            }
            throw new SyncSourceException(e);
        }

        for (int i = 0; i < allExcContacts.length; i++) {
            try {
                cm.removeContact(allExcContacts[i], this.username, this.credentials, this.exchangeFolder);
            }
            catch (DataAccessException e) {
                if (log.isLoggable(Level.SEVERE)) {
                    log.log(Level.SEVERE, "Error removing contact with id " +
                            allExcContacts[i].getId() + " from Exchange.");
                }
                if (log.isLoggable(Level.INFO)) {
                    log.log(Level.INFO, e.getMessage(), e);
                }

            }
        }

        if (log.isLoggable(Level.INFO)) {
            log.log(Level.INFO, "Clearing cached contacts for source uri " +
                    this.sourceURI + " and principal " + this.principalId);
        }
        try {
            new ItemManager().deleteItems(cm.getAllItems(this.sourceURI, this.principalId),
                    this.sourceURI, this.principalId);
        }
        catch (DataAccessException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.log(Level.SEVERE, "Error while clearing cached contacts: " + e.getMessage() + ".");
            }
            if (log.isLoggable(Level.INFO)) {
                log.log(Level.INFO, e.getMessage(), e);
            }
        }

    }

    // --------------------------------------------------------- Private methods


    private Contact getContact(String id,
                               String content,
                               Date t)
            throws SyncSourceException {

        Contact c = new Contact(id);

        c.setLastUpdate(t);

        if (content != null && content.length() > 0) {

            if (TYPE_VCARD.equals(getType())) {
                vcard2Contact(content, c);
            } else {
                sifc2Contact(content, c);
            }
        }

        return c;

    }

    /**
     * Create SyncItem array from Contact array
     *
     * @param contacts
     * @return array of SyncItem
     * @throws SyncSourceException
     */
    private SyncItem[] getSyncItems(Contact[] contacts)
            throws SyncSourceException {

        SyncItem[] syncItems;

        syncItems = new SyncItem[contacts.length];

        for (int i = 0, l = syncItems.length; i < l; i++) {
            syncItems[i] = getSyncItem(contacts[i]);
        }

        return syncItems;

    }

    /**
     * Create SyncItem from Contact
     *
     * @param contact
     * @return SyncItem
     * @throws SyncSourceException
     */
    private SyncItem getSyncItem(Contact contact)
            throws SyncSourceException {

        SyncItem syncItem;
        String content;

        syncItem = new SyncItemImpl(this,
                contact.getId(),
                contact.getStatus());

        if (TYPE_VCARD.equals(getType())) {
            content = contact2vcard(contact);
        } else {
            content = contact2sifc(contact);
        }

        syncItem.setContent(content.getBytes());

        syncItem.setType(this.getType());

        return syncItem;

    }

    /**
     * Create Item array from Contact array
     *
     * @param contacts
     * @return array of SyncItem
     */
    private Item[] getItems(Contact[] contacts) {

        Item[] items;

        items = new Item[contacts.length];

        for (int i = 0, l = items.length; i < l; i++) {
            items[i] = getItem(contacts[i]);
        }

        return items;

    }

    /**
     * Create Item from Contact
     *
     * @param contact
     * @return item
     */
    private Item getItem(Contact contact) {

        Item item;

        String firstName = null;
        String lastName = null;

        Name name = contact.getName();

        if (name != null) {
            if (name.getFirstName() != null) {
                firstName = name.getFirstName().
                        getPropertyValueAsString();
            }
            if (name.getLastName() != null) {
                lastName = name.getLastName().
                        getPropertyValueAsString();
            }
        }

        long crc = StringTools.getCRCFromString(firstName + lastName);

        item = new Item(contact.getId(),
                contact.getHref(),
                contact.getLastUpdate(),
                crc);

        return item;

    }

    /**
     * Converts the given contact into a vcard String
     *
     * @param c the contact to convert
     * @return the vcard
     * @throws SyncSourceException in case of convertion errors
     */
    private String contact2vcard(Contact c)
            throws SyncSourceException {
        try {
            return new ContactToVcard(deviceTimeZone, deviceCharset).convert(c);
        } catch (ConverterException e) {
            throw new SyncSourceException("Conversion error for item "
                    + c.getId()
                    + ": "
                    + e.getMessage()
                    , e
            );
        }
    }

    /**
     * Converts the given contact into a sifc String
     *
     * @param c the contact to convert
     * @return the sifc document
     * @throws SyncSourceException in case of convertion errors
     */
    private String contact2sifc(Contact c)
            throws SyncSourceException {
        String sifc;
        try {
            sifc = new ContactToXML(deviceTimeZone, deviceCharset).convert(c);
        } catch (ConverterException e) {
            throw new SyncSourceException("Convertion error for item "
                    + c.getId()
                    + ": "
                    + e.getMessage()
                    , e
            );
        }
        return sifc;
    }

    /**
     * Fill the given Contact object with the properties in the given vcard
     *
     * @param content message in VCard
     * @param c       the contect to feed
     * @throws SyncSourceException
     */
    private void vcard2Contact(String content, Contact c)
            throws SyncSourceException {

        //
        // just make sure we have something to do...
        //
        if ((content == null) || (content.length() == 0)) {
            return;
        }

        ByteArrayInputStream is = null;
        VcardParser parser;

        com.funambol.foundation.pdi.contact.Contact contact;

        try {

            /**
             * This is used to replace the String "=\r\n" with the
             * String "\r\n ": This replace is necessary because some devices do
             * not send the correct line delimiter for long lines.
             */
            //content = content.replaceAll("=\r\n","\r\n ");
            content = SourceUtils.handleLineDelimiting(content);


            is = new ByteArrayInputStream(content.getBytes());

            //converting the vcard into a Contact object

            parser = new VcardParser(is, deviceTimeZoneDescr, deviceCharset);

            parser.setLogger(log);

            contact = parser.vCard();

            //
            // We now have to convert a portal contact into a foundation contact
            //
            c.setUid(contact.getUid());
            c.setTimezone(contact.getTimezone());
            c.setNotes(contact.getNotes());
            c.setRevision(contact.getRevision());
            c.setName(contact.getName());
            c.setBusinessDetail(contact.getBusinessDetail());
            c.setPersonalDetail(contact.getPersonalDetail());
            c.setLanguages(contact.getLanguages());
            c.setCategories(contact.getCategories());
            c.setFolder(contact.getFolder());
            c.setSubject(contact.getSubject());
            c.setMileage(contact.getMileage());

            c.setSensitivity(contact.getSensitivity());
            c.setImportance(contact.getImportance());

        } catch (Exception e) {
            e.printStackTrace();
            throw new SyncSourceException("Convertion error for item "
                    + c.getId()
                    + ": "
                    + e.getMessage()
                    , e
            );
        } finally {
            if (is != null) {
                try {
                    is.close();
                }
                catch (Exception e) {
                    if (log.isLoggable(Level.ALL)) {
                        log.log(Level.ALL, e.getMessage());
                    }
                }
            }
        }
    }

    /**
     * Fill the given Contact object with the properties in the given sifc document
     *
     * @param content message in VCard
     * @param c       the contect to feed
     * @throws SyncSourceException
     */
    private void sifc2Contact(String content, Contact c)
            throws SyncSourceException {
        //
        // just make sure we have something to do...
        //
        if ((content == null) || (content.length() == 0)) {
            return;
        }


        ByteArrayInputStream is = null;
        XMLContactParser parser;

        com.funambol.foundation.pdi.contact.Contact contact;

        try {
            log.finest("SIF-C to convert to Contact:\n" + content);
            is = new ByteArrayInputStream(content.getBytes());

            //converting the vcard into a Contact object

            parser = new XMLContactParser(is);

            contact = parser.parse();

            //
            // We now have to convert a portal contact into a foundation contact
            //
            c.setUid(contact.getUid());
            c.setTimezone(contact.getTimezone());
            c.setNotes(contact.getNotes());
            c.setRevision(contact.getRevision());
            c.setName(contact.getName());
            c.setBusinessDetail(contact.getBusinessDetail());
            c.setPersonalDetail(contact.getPersonalDetail());
            c.setLanguages(contact.getLanguages());
            c.setCategories(contact.getCategories());
            c.setFolder(contact.getFolder());
            c.setSubject(contact.getSubject());
            c.setMileage(contact.getMileage());
            c.setSensitivity(contact.getSensitivity());
            c.setImportance(contact.getImportance());

        } catch (Exception e) {
            throw new SyncSourceException("Parsing error for item "
                    + c.getId()
                    + ": "
                    + e.getMessage()
                    , e
            );
        } finally {
            if (is != null) {
                try {
                    is.close();
                }
                catch (Exception e) {
                    if (log.isLoggable(Level.ALL)) {
                        log.log(Level.ALL, e.getMessage());
                    }
                }
            }
        }
    }

    /**
     * Creates a new Contact from the given SyncItem
     *
     * @param syncItem SyncItem
     * @return Contact
     * @throws SyncSourceException
     */
    private Contact getContactFromSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        Contact contact;
        String itemKey = syncItem.getKey().getKeyAsString();

        byte[] itemContent = syncItem.getContent();

        if (itemContent == null) {
            itemContent = new byte[0];
        }

        String content = new String(itemContent);

        Date t = syncItem.getTimestamp();

        contact = getContact(itemKey, content, t);

        return contact;
    }

    /**
     *
     */
    private void removeSyncItem(SyncItemKey itemKey)
            throws SyncSourceException {

        Contact contact;

        String id;
        String href;

        id = itemKey.getKeyAsString();

        href = getHref(id, principalId);

        try {
            if (href != null) {
                contact = new Contact(id);
                contact.setHref(href);

                this.cm.removeContact(contact,
                        username,
                        credentials,
                        exchangeFolder);
                toSyncRemoveItems(contact.getId());
                removeInExchangeItemsKeys(contact.getId());
            } else {
                if (log.isLoggable(Level.SEVERE)) {
                    log.severe
                            ("Remove contact failed. Contact with " + id + " not found while reading fnbl_exch_cache");
                }
                toSyncErrorRemoveItems(id);
            }
        } catch (Exception e) {

            toSyncErrorRemoveItems(itemKey.getKeyAsString());

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "removeSyncItem", e);
            }

            throw new SyncSourceException("Error removing items", e);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Exchange SyncSource " +
                    sourceURI +
                    " - removing sync item");
        }

    }

    /**
     * <p>Updates the SyncItem with the the input key from the SyncSource's cache
     *  with the values from the input Contacts and sent the input timezone for it.</p>
     * @param chachedItemKey (in) key of the Sync Item in the SyncSource's cache
     * @param contactToSet (in) Contact to set in the sync item from cache
     * @param timeToSet the timezone to set for the sync item from cache
     * @throws SyncSourceException is an error occurs
     */
    private void updateContactInCache(String chachedItemKey, Contact contactToSet, Timestamp timeToSet)
            throws SyncSourceException {
        List cachedSyncItems;
        Iterator iterator;
        SyncItem syncItem;
        String content, key;

        cachedSyncItems = this.getCachedSyncItems();

        iterator = cachedSyncItems.iterator();

        while (iterator.hasNext()) {
            syncItem = (SyncItem) iterator.next();
            if (syncItem == null || syncItem.getKey() == null) {
                continue;
            }

            key = syncItem.getKey().getKeyAsString();

            if (key.equals(chachedItemKey)) {

                if (TYPE_VCARD.equals(getType())) {
                    content = contact2vcard(contactToSet);
                } else {
                    content = contact2sifc(contactToSet);
                }

                syncItem.setContent(content.getBytes());
                syncItem.setTimestamp(timeToSet);
                break;
            }
        }

    }

}
