/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the Honest Public License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Honest Public License for more details.
 *
 * You should have received a copy of the Honest Public License
 * along with this program; if not, write to Funambol,
 * 643 Bair Island Road, Suite 305 - Redwood City, CA 94063, USA
 */
package com.funambol.exchange.engine.source;

import java.io.Serializable;

import java.sql.Timestamp;

import java.util.Date;
import java.util.logging.Logger;
import java.util.logging.Level;

import com.funambol.framework.core.AlertCode;
import com.funambol.framework.engine.SyncItemImpl;
import com.funambol.framework.engine.SyncItem;
import com.funambol.framework.engine.SyncItemKey;
import com.funambol.framework.engine.source.SyncSource;
import com.funambol.framework.engine.source.SyncSourceException;
import com.funambol.framework.server.store.NotFoundException;
import com.funambol.framework.logging.Sync4jLogger;

import com.funambol.exchange.DataAccessException;
import com.funambol.exchange.items.common.model.Item;
import com.funambol.exchange.items.common.manager.ItemManager;
import com.funambol.exchange.items.note.model.Note;
import com.funambol.exchange.items.note.manager.NoteManager;
import com.funambol.exchange.items.note.NoteParseException;
import com.funambol.exchange.util.StringTools;
import com.funambol.exchange.util.Def;

/**
 * This class define <i>SyncSource</i>
 * between SyncServer and Exchange Server notes items
 *
 * @version $Id: ExchangeNoteSyncSource.java,v 1.1 2006/08/08 07:52:15 gibi Exp $
 */
public class ExchangeNoteSyncSource
        extends ExchangeSyncSource
        implements SyncSource, Serializable {

    // --------------------------------------------------------------- Constants

    private static final String EXCHANGE_HREF_EXTENSION = ".eml";

    // ------------------------------------------------------------ Private data

    private NoteManager nm = null;

    // ------------------------------------------------------------ Constructors

    public ExchangeNoteSyncSource() {
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * init manager
     */
    public void initManager()
            throws SyncSourceException {

        try {

            this.nm = new NoteManager(this.host,
                    this.port,
                    this.exchangeHost);

        } catch (Exception e) {

            throw new SyncSourceException("Error init manager: " +
                    e.getMessage(), e);
        }

    }

    /**
     * @see SyncSource
     */
    public SyncItem addSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        Note note;
        Note newNote = null;

        String itemKey;
        String href;

        try {

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // created
            // default href = this.CurrentTimesMillis()
            //
            href = String.valueOf(System.currentTimeMillis()) +
                    EXCHANGE_HREF_EXTENSION;

            if (log.isLoggable(Level.FINEST)) {
                log.finest(itemKey + " is new: adding");
            }

            note = getNoteFromSyncItem(syncItem);

            note.setHref(href);

            newNote = this.nm.addNote(note,
                    username,
                    credentials,
                    exchangeFolder);

            long crc = StringTools.getCRCFromString(newNote.getTextDescription());

            toSyncNewItems(newNote.getId(), crc, newNote.getLastModified());
            addInExchangeItemsKeys(newNote.getId());

            syncItem = getSyncItem(newNote);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Exchange SyncSource " +
                        sourceURI +
                        " - set syncItem" +
                        syncItem.getKey().getKeyAsString());
            }

            return syncItem;

        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "addSyncItem", e);
            }

            if (newNote != null) {
                toSyncErrorNewItems(newNote.getId());
            }

            throw new SyncSourceException("Error setting the item " +
                    syncItem + " " + e.getMessage(), e);
        }
    }

    /**
     * @see SyncSource
     */
    public SyncItem updateSyncItem(SyncItem syncItem)
            throws SyncSourceException {

        Note note;
        Note updateNote = null;

        String itemKey;
        String href;

        try {

            itemKey = syncItem.getKey().getKeyAsString();

            //
            // create href [eg. 1234.eml] in exchange items
            //
            href = getHref(itemKey, principalId);

            if (log.isLoggable(Level.FINEST)) {
                log.finest(itemKey + " already exists: updating");
            }

            note = getNoteFromSyncItem(syncItem);

            note.setHref(href);

            updateNote = this.nm.updateNote(note,
                    username,
                    credentials,
                    exchangeFolder);

            long crc = StringTools.getCRCFromString(updateNote.getTextDescription());

            toSyncUpdateItems(updateNote.getId(), crc, updateNote.getLastModified());

            syncItem = getSyncItem(updateNote);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Exchange SyncSource " +
                        sourceURI +
                        " - set syncItem" +
                        syncItem.getKey().getKeyAsString());
            }

            return syncItem;


        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "setSyncItem", e);
            }

            if (updateNote != null) {
                toSyncErrorUpdateItems(updateNote.getId());
            }

            throw new SyncSourceException("Error setting the item " +
                    syncItem + " " + e.getMessage(), e);
        }
    }

    /**
     * @see SyncSource
     */
    public SyncItem getSyncItemFromId(SyncItemKey syncItemKey)
            throws SyncSourceException {

        Note note = null;
        String id;
        SyncItem syncItem;

        id = syncItemKey.getKeyAsString();

        //
        // search syncItem in cache
        //
        syncItem = getSyncItemById(id);

        if (syncItem != null) {
            return syncItem;
        }

        if (log.isLoggable(Level.SEVERE)) {
            log.severe
                    ("Note with id " + id + " not found while reading Exchange cache");
        }

        try {

            note = this.nm.getNoteById(username,
                    credentials,
                    id,
                    this.exchangeFolder);
        } catch (Exception e) {

            Throwable previous = e.getCause();

            if (previous instanceof NotFoundException) {

                if (log.isLoggable(Level.SEVERE)) {
                    log.severe
                            ("Note not found while reading Exchange database: " +
                                    e.getMessage());
                }

                Sync4jLogger.getLogger(Def.LOGGER_NAME).
                        throwing(getClass().
                                getName(), "readExchangeDatabase", e);

            } else {

                if (log.isLoggable(Level.FINEST)) {
                    log.severe("SyncSource error: " + e.getMessage());
                    log.throwing(getClass().getName(), "getSyncItemFromId", e);
                }

                throw new SyncSourceException("Error reading items", e);

            }
        }

        if (note != null) {
            note.setState(getSyncItemStatus(note.getId()));
            syncItem = getSyncItem(note);
        }

        return syncItem;
    }

    /**
     * @see SyncSource
     */
    public void removeSyncItem(SyncItemKey itemKey, Timestamp time, boolean softDelete)
            throws SyncSourceException {

        if (!softDelete) {
            removeSyncItem(itemKey);
        }

    }

    /**
     * @see SyncSource
     */
    public SyncItemKey[] getSyncItemKeysFromTwin(SyncItem syncItem)
            throws SyncSourceException {

        Item[] items;

        try {

            Note note = getNoteFromSyncItem(syncItem);
            items = this.nm.getTwinItems(note,
                    principalId,
                    this.sourceURI);
        } catch (Exception e) {

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "getSyncItemKeysFromTwin", e);
            }

            throw new SyncSourceException("Error getting twin", e);
        }

        return getSyncItemKeysFromItems(items);
    }

    /**
     * @param username
     * @param credentials
     * @param exchangeFolder
     * @return items array
     */
    public Item[] getExchangeItems(String username,
                                   String credentials,
                                   String exchangeFolder,
                                   String[] ids,
                                   boolean isBeginSync)
            throws SyncSourceException {

        Item  [] items;
        Note  [] notes;

        try {

            //
            // is not slow sync or called from endSync
            // load calendar with only fields for twin setting
            //
            if (((syncMode != AlertCode.SLOW) &&
                    (syncMode != AlertCode.REFRESH_FROM_SERVER)) ||
                    !isBeginSync) {

                notes = this.nm.getNotesForTwin(username,
                        credentials,
                        ids,
                        exchangeFolder);


            } else {

                notes = this.nm.getNotes(username,
                        credentials,
                        ids,
                        exchangeFolder);

                addToCacheSyncItems(getSyncItems(notes));

            }

            items = getItems(notes);


        } catch (DataAccessException e) {
            throw new SyncSourceException("Error getting items", e);
        }

        return items;

    }

    /**
     * @param username
     * @param credentials
     * @param ids
     * @param exchangeFolder
     * @return SyncItem array
     */
    public SyncItem[] getSyncItems(String username,
                                   String credentials,
                                   String[] ids,
                                   String exchangeFolder)
            throws SyncSourceException {

        SyncItem [] syncItems;
        Note     [] notes;

        try {

            notes = this.nm.getNotes(username,
                    credentials,
                    ids,
                    exchangeFolder);

            syncItems = getSyncItems(notes);


        } catch (DataAccessException e) {
            throw new SyncSourceException("Error getting items", e);
        }

        return syncItems;

    }

    /**
     * Delete all notes from the Exchange server
     *
     * @see ExchangeSyncSource
     */
    public void deleteAll()
            throws SyncSourceException {
        if (log.isLoggable(Level.INFO)) {
            log.log(Level.INFO, "Removing all notes from Exchange for user " + this.username + ".");
        }

        Note[] allExcNotes;
        try {
            allExcNotes = nm.getNotes(this.username, this.credentials, new String[0], this.getExchangeFolder());
        }
        catch (DataAccessException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.log(Level.SEVERE, "Error removing all notes from Exchange.");
            }
            throw new SyncSourceException(e);
        }

        for (int i = 0; i < allExcNotes.length; i++) {
            try {
                nm.removeNote(allExcNotes[i], this.username, this.credentials, this.exchangeFolder);
            }
            catch (DataAccessException e) {
                if (log.isLoggable(Level.SEVERE)) {
                    log.log(Level.SEVERE, "Error removing note with id " +
                            allExcNotes[i].getId() + " from Exchange.");
                }
                if (log.isLoggable(Level.INFO)) {
                    log.log(Level.INFO, e.getMessage(), e);
                }

            }
        }

        if (log.isLoggable(Level.INFO)) {
            log.log(Level.INFO, "Clearing cached notes for source uri " +
                    this.sourceURI + " and principal " + this.principalId);
        }
        try {
            new ItemManager().deleteItems(nm.getAllItems(this.sourceURI, this.principalId),
                    this.sourceURI, this.principalId);
        }
        catch (DataAccessException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.log(Level.SEVERE, "Error while clearing cached notes: " + e.getMessage() + ".");
            }
            if (log.isLoggable(Level.INFO)) {
                log.log(Level.INFO, e.getMessage(), e);
            }
        }
    }

    // --------------------------------------------------------- Private methods

    /**
     * Create SyncItem array from Contact array
     *
     * @param notes
     * @return array of SyncItem
     * @throws SyncSourceException
     */
    private SyncItem[] getSyncItems(Note[] notes)
            throws SyncSourceException {

        SyncItem[] syncItems;

        syncItems = new SyncItem[notes.length];

        for (int i = 0, l = syncItems.length; i < l; i++) {
            syncItems[i] = getSyncItem(notes[i]);
        }

        return syncItems;

    }

    /**
     * Create SyncItem from Note
     *
     * @param note
     * @return SyncItem
     * @throws SyncSourceException
     */
    private SyncItem getSyncItem(Note note)
            throws SyncSourceException {

        SyncItem syncItem;

        syncItem = new SyncItemImpl(this,
                note.getId(),
                note.getState());

        try {

            syncItem.setContent(note.toXML().getBytes());

            syncItem.setType(this.getType());

        } catch (NoteParseException e) {
            throw new SyncSourceException("Error setting the item "
                    + syncItem
                    , e
            );
        }

        return syncItem;

    }

    /**
     * Create Item array from Note array
     *
     * @param notes
     * @return array of SyncItem
     */
    private Item[] getItems(Note[] notes) {

        Item[] items;

        items = new Item[notes.length];

        for (int i = 0, l = items.length; i < l; i++) {
            items[i] = getItem(notes[i]);
        }

        return items;

    }

    /**
     * Create Item from Note
     *
     * @param note
     * @return item
     */
    private Item getItem(Note note) {

        Item item;
        String noteBody;

        if (note.getTextDescription() != null) {
            noteBody = note.getTextDescription();
        }
        else {
            noteBody = "";
        }

        item = new Item(note.getId(),
                note.getHref(),
                note.getLastModified(),
                StringTools.getCRCFromString(noteBody));
        return item;

    }

    /**
     * Creates a new Note from the given syncItem
     *
     * @param syncItem SyncItem
     * @return Note
     */
    private Note getNoteFromSyncItem(SyncItem syncItem)
            throws Exception {

        byte[] itemContent = syncItem.getContent();

        if (itemContent == null) {
            itemContent = new byte[0];
        }

        String content = new String(itemContent);

        Date t = syncItem.getTimestamp();

        return new Note(syncItem.getKey().getKeyAsString(),
                content,
                t);
    }

    /**
     *
     */
    private void removeSyncItem(SyncItemKey itemKey)
            throws SyncSourceException {

        Note note;

        String id;
        String href;

        id = itemKey.getKeyAsString();

        href = getHref(id, principalId);

        try {
            if (href != null) {
                note = new Note(id);
                note.setHref(href);

                this.nm.removeNote(note,
                        username,
                        credentials,
                        exchangeFolder);
                toSyncRemoveItems(note.getId());
                removeInExchangeItemsKeys(note.getId());
            } else {
                if (log.isLoggable(Level.SEVERE)) {
                    log.severe
                            ("Remove note failed. Note with " + id + " not found while reading fnbl_exch_cache");
                }
                toSyncErrorRemoveItems(id);
            }

        } catch (Exception e) {
            toSyncErrorRemoveItems(itemKey.getKeyAsString());

            if (log.isLoggable(Level.FINEST)) {
                log.severe("SyncSource error: " + e.getMessage());
                log.throwing(getClass().getName(), "removeSyncItem", e);
            }

            throw new SyncSourceException("Error remove items", e);
        }

    }

}
